(* $Id: godi_exec.ml 545 2005-04-04 18:24:48Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)


(* This implementation uses the "shell" library, because it has job
 * control (commands run in their own process group).
 *
 * It does not contain special support for:
 * - cd to directory
 * - run callbacks periodically (this is here done with an itimer)
 *)

type process_status = Unix.process_status =
    WEXITED of int 
  | WSIGNALED of int
  | WSTOPPED of int

type collector =
    [ `Buffer of Buffer.t
    | `Forward of (string -> unit)
    ]

type emitter =
    [ `File of string
    ]

type interactivity =
    { period : float;
      callback : unit -> bool;
    }

exception Command_not_found of string
exception Command_failure of (string * string array * process_status)
exception Command_discontinued


let collect collector s =
  match collector with
      `Buffer b -> Buffer.add_string b s
    | `Forward f ->
	try
	  f s
	with
	    error ->
	      prerr_endline ("Async exception: " ^ Printexc.to_string error)
;;


let collect_buf strbuf collector fd =
  let n =
    try
      Unix.read fd strbuf 0 (String.length strbuf)
    with
	Unix.Unix_error(Unix.EAGAIN,_,_) -> (-1)
      | Unix.Unix_error(Unix.EWOULDBLOCK,_,_) -> (-1)
  in
  let s = String.sub strbuf 0 (max n 0) in
  collect collector s;
  if n = 0 then Unix.close fd;
  (n<>0)
;;


let call_cmd_1
      ?(env = [])
      ?stdin
      ?stdout
      ?stderr
      ?interactivity
      (cmd, args) =

  let cleanup = ref [] in

  try
    let ia_handler = ref (fun () -> ()) in

    let e = Shell_sys.current_env() in
    List.iter (fun (n,v) -> Shell_sys.set_env_var e n v) env;
    let sh_cmd = Shell_sys.command
		   ~arguments:args
		   ~environment:e
		   ~filename:(Shell_sys.lookup_executable cmd) () in
    let stdin_producer =
      match stdin with
	  None -> 
	    Some(Shell.from_dev_null)
	| Some (`File fn) ->
	    let fd = Unix.openfile fn [Unix.O_RDONLY] 0 in
	    let fd_ref = ref (Some fd) in
	    cleanup := (fun () -> 
			  match !fd_ref with
			      Some fd -> Unix.close fd; fd_ref := None
			    | None -> ()) :: !cleanup;
	    Some(Shell.from_fd fd)
    in
    let stdout_consumer =
      match stdout with
	  None -> None
	| Some outcoll ->
	    let strbuf = String.create 1024 in
	    Some(Shell.to_function
		   ~consumer:(fun fd ->
				let r = collect_buf strbuf outcoll fd in
				!ia_handler();
				r
			     )
		   ())
    in
    let stderr_consumer =
      match stderr with
	  None -> None
	| Some errcoll ->
	    let strbuf = String.create 1024 in
	    Some(Shell.to_function
		   ~consumer:(fun fd ->
				let r = collect_buf strbuf errcoll fd in
				!ia_handler();
				r
			     )
		   ())
    in
    let ues = Unixqueue.create_unix_event_system() in
    let cmd_eng =
      new Shell_uq.call_engine 
	~mode:Shell_sys.Background 
	?stdin:stdin_producer
	?stdout:stdout_consumer 
	?stderr:stderr_consumer
	[ sh_cmd ] ues in

    let aborted = ref false in
    let timer_eng = ref None in

    ( match interactivity with
	  None -> ()
	| Some ia ->
	    ia_handler := ( 
	      fun () ->
		(* Check interactivity: *)
		let continue =
		  try
		    ia.callback()
		  with error ->
		    prerr_endline("Async exception: " ^ Printexc.to_string error);
		    true
		in
		if not continue then (
		  (* Kill the process group. Note that we do not call
		   * Shell_sys.abandon_job, because there are problems in the
		   * Shell_sys module (race conditions?)
		   *)
		  if not !aborted then
		    Shell_sys.kill_process_group cmd_eng#job_instance;
		  aborted := true
		)
	    );
	    (* Timer implementation: The poll_engine sleeps the period of time,
	     * and restarts when the period is over.
	     *)
	    let wid = Unixqueue.new_wait_id ues in
	    let eng =
	      new Uq_engines.poll_engine
		[ Unixqueue.Wait wid, ia.period ] ues in
	    timer_eng := Some eng;
	    Uq_engines.when_state
	      ~is_done:(fun _ -> !ia_handler(); eng#restart()) eng;
    );

    (* When cmd_eng is done, also cancel timer_eng: *)
    let cancel_timer() =
      match !timer_eng with
	  None -> ()
	| Some e -> e # abort()
    in
    Uq_engines.when_state 
      ~is_done:(fun _ -> cancel_timer())
      ~is_error:(fun _ -> cancel_timer())
      cmd_eng;

    (* Wait until the engines are done: *)
    Unixqueue.run ues;

    List.iter (fun f -> f()) !cleanup;
    cleanup := [];

    (* Check engine status: *)
    match cmd_eng # state with
	`Done _ ->
	  ()
      | `Error e ->
	  if !aborted then
	    raise Command_discontinued
	  else
	    raise e
      | `Aborted | `Working _ ->
	  ()
  with
      error ->
	List.iter (fun f -> f()) !cleanup;
	raise error
;;


(* Old implementation with signal-based timers: *)

(*
let call_cmd_1
      ?(env = [])
      ?stdin
      ?stdout
      ?stderr
      ?interactivity
      (cmd, args) =

  let cleanup = ref [] in

  try
    let ia_handler = ref (fun () -> ()) in

    let e = Shell_sys.current_env() in
    List.iter (fun (n,v) -> Shell_sys.set_env_var e n v) env;
    let sh_cmd = Shell_sys.command
		   ~arguments:args
		   ~environment:e
		   ~filename:(Shell_sys.lookup_executable cmd) () in
    let job = Shell_sys.new_job () in
    Shell_sys.add_command sh_cmd job;
    ( match stdin with
	  None -> 
	    let producer = (fun fd -> Unix.close fd; false) in
	    Shell_sys.add_producer ~producer sh_cmd job;
	    (* Close stdin immediately (same as assigning /dev/null) *)
	| Some (`File fn) ->
	    let fd = Unix.openfile fn [Unix.O_RDONLY] 0 in
	    Shell_sys.set_assignments sh_cmd [ fd, Unix.stdin ];
	    let fd_ref = ref (Some fd) in
	    cleanup := (fun () -> 
			  match !fd_ref with
			      Some fd -> Unix.close fd; fd_ref := None
			    | None -> ()) :: !cleanup
    );
    ( match stdout with
	  None -> ()
	| Some outcoll ->
	    let strbuf = String.create 1024 in
	    Shell_sys.add_consumer
	      ~descr:Unix.stdout
	      ~consumer:(fun fd ->
			   let r = collect_buf strbuf outcoll fd in
			   !ia_handler();
			   r
			)
	      sh_cmd
	      job
    );
    ( match stderr with
	  None -> ()
	| Some errcoll ->
	    let strbuf = String.create 1024 in
	    Shell_sys.add_consumer
	      ~descr:Unix.stderr
	      ~consumer:(fun fd ->
			   let r = collect_buf strbuf errcoll fd in
			   !ia_handler();
			   r
			)
	      sh_cmd
	      job
    );
    let instance =
      Shell_sys.run_job ~mode:Shell_sys.Background job in
    
    let aborted =
      ref false in
    
    ( match interactivity with
	  None -> ()
	| Some ia ->
	    ia_handler := fun () ->
	      (* Check interactivity: *)
	      let continue =
		try
		  ia.callback()
		with error ->
		  prerr_endline("Async exception: " ^ Printexc.to_string error);
		  true
	      in
	      if not continue then (
		(* Kill the process group. Note that we do not call
		 * Shell_sys.abandon_job, because there are problems in the
		 * Shell_sys module (race conditions?)
		 *)
		if not !aborted then
		  Shell_sys.kill_process_group instance;
		aborted := true
	      )
    );
    
    (* Set Timer such that Shell_sys.finish_job is interrupted by EINTR
     * every time the timer expires
     *)
    let n =
      match interactivity with
	  None -> 0.0
	| Some ia -> ia.period in
    ignore(
      Unix.setitimer
	     Unix.ITIMER_REAL
	     { Unix.it_interval = n;
	       Unix.it_value = n;
	     });
    cleanup := (fun () ->
		  (* Stop timer: *)
		  ignore(
		    Unix.setitimer
			   Unix.ITIMER_REAL
			   { Unix.it_interval = 0.0;
			     Unix.it_value = 0.0;
			   })) :: !cleanup;
    
    (* Loop until the command has finished: *)
    let finished = ref false in
    while not !finished do
      try
	Shell_sys.finish_job instance;
	finished := true
      with
	  Unix.Unix_error(Unix.EINTR,_,_) ->
	    !ia_handler()
    done;

    List.iter (fun f -> f()) !cleanup;
    cleanup := [];

    (* Check job status: *)
    match Shell_sys.job_status instance with
	Shell_sys.Job_running -> assert false
      | Shell_sys.Job_partially_running -> assert false
      | Shell_sys.Job_ok -> ()
      | Shell_sys.Job_error ->
	  if !aborted then
	    raise Command_discontinued;
	  ( match Shell_sys.processes instance with
		[ p ] ->
		  let pstatus = Shell_sys.status p in
		  raise(Command_failure(cmd,args,pstatus))
	      | _ -> assert false
	  )
      | Shell_sys.Job_abandoned ->
	  raise Command_discontinued
  with
      error ->
	List.iter (fun f -> f()) !cleanup
;;
*)


let call_cmd
      ?env
      ?cd
      ?stdin
      ?stdout
      ?stderr
      ?interactivity
      (cmd, args) =
  (* Remember the old directory, and go back to it after the command
   * has finished.
   * This is not the best way to implement the ~cd feature. It would be
   * better the "shell" module would allow us to chdir in the already
   * forked subprocess.
   *)
  let cwd = Unix.getcwd() in
  try
    ( match cd with
	  None -> ()
	| Some dir -> Unix.chdir dir
    );
    call_cmd_1 ?env ?stdin ?stdout ?stderr ?interactivity (cmd, args);
    Unix.chdir cwd
  with
    | Shell_sys.Executable_not_found name ->
	Unix.chdir cwd;
	raise(Command_not_found name)
    | error ->
	Unix.chdir cwd;
	raise error
;;


let shell cmd =
  ("/bin/sh", [| "-c"; cmd |])
;;


type signal_handler_id = int;;


let signal_handlers = ref [];;
let signal_id = ref 0;;

let handle_signal signo =
  let break = ref false in
  List.iter
    (fun (_,n,f) ->
       if n = signo then (
	 try
	   let () = f signo in ()
	 with
	   | Sys.Break ->
	       break := true
	   | error ->
	       prerr_endline("Exception in signal handler: " ^
			     Printexc.to_string error)
       )
    )
    !signal_handlers;
  if !break then raise Sys.Break
;;


let register_signal_handler signo f =
  let id = !signal_id in
  incr signal_id;
  if not(List.exists (fun (_,n,_) -> n = signo) !signal_handlers) then
    Sys.set_signal signo (Sys.Signal_handle handle_signal);
  signal_handlers := (id, signo, f) :: !signal_handlers;
  id
;;


let forget_signal_handler id =
  try
    let (_,signo,_) = List.find (fun (x,_,_) -> x = id) !signal_handlers in
    signal_handlers := List.filter (fun (x,_,_) -> x <> id) !signal_handlers;
    if not(List.exists (fun (_,n,_) -> n = signo) !signal_handlers) then
      Sys.set_signal signo Sys.Signal_default;
  with
      Not_found ->
	failwith ("forget_signal_handler")
;;


let environment = Unix.environment;;


let set_path p =
  Unix.putenv "PATH" p
;;


let set_variable = Unix.putenv
;;


(* Initialization: *)
(* Don't install any handlers, we are installing them ourselves: *)
Shell_sys.configure_job_handlers
  ~catch_sigint:false
  ~catch_sigquit:false
  ~catch_sigterm:false
  ~catch_sighup:false
  ~catch_sigchld:false
  ~set_sigpipe:false
  ~at_exit:false
  ();
Shell_sys.install_job_handlers ()
;;


let forward signo =
  Shell_sys.iter_job_instances
    (fun job ->
       try
	 Shell_sys.kill_process_group ~signal:signo job
       with
	   Unix.Unix_error(Unix.ESRCH,_,_) ->
	     ())
;;


let forward_and_break signo =
  forward signo;
  raise Sys.Break
;;


(* Set locale variables to avoid problems with exotic character sets *)
let _ =
  Unix.putenv "LC_CTYPE" "C";
  Unix.putenv "LC_COLLATE" "C"
;;

(* Install our own signal handlers: *)
(* - Ensure there is a SIGALRM handler, even if empty *)
let _ = register_signal_handler Sys.sigalrm (fun _ -> ()) in
(* - Forward these signals to subprocesses: *)
let _ = register_signal_handler Sys.sigint forward_and_break in
let _ = register_signal_handler Sys.sigquit forward_and_break in
let _ = register_signal_handler Sys.sigterm forward_and_break in
let _ = register_signal_handler Sys.sighup forward_and_break in
let _ = register_signal_handler Sys.sigcont forward in
(* - SIGTSTP must be forwarded, but we must also stop this process: *)
let _ = register_signal_handler Sys.sigtstp
	  (fun signo ->
	     forward signo;
	     Unix.kill (Unix.getpid()) Sys.sigstop) in
(* - Ignore SIGPIPE, we have exception handling: *)
let _ = register_signal_handler Sys.sigpipe (fun _ -> ()) in
(* - SIGCHLD must traverse the children: *)
let _ = register_signal_handler Sys.sigchld
	  (fun _ -> Shell_sys.watch_for_zombies()) in
(* atexit: terminate the subprocesses *)
at_exit (fun _ -> forward Sys.sigterm);
();;
